/*! \file nanocrypt.c
    \brief Nanocrypt program source file
	 \author Damien Miller <djm@mindrot.org>
	 \version 0.0.0
	 \date 1999
	 
	 A small program to encrypt and decrypt files using the RC4 
	 algroithm.

    This code is licensed under the GNU GPL. Please see the file
    COPYING for details
	 
	 \warning The name "RC4" is trademarked in the United States, 
	 \warning you may need to use "RC4 compatible" or "ARC4" 
	 \warning (Alleged RC4).
*/

/* $Id: nanocrypt.c,v 1.3 1999/10/03 01:05:29 dmiller Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "rc4.h"

#ifndef PROGRAM_NAME
#define PROGRAM_NAME "nanocrypt"
#endif /* PROGRAM_NAME */

#ifndef PROGRAM_VERSION
#define PROGRAM_VERSION "unspecified"
#endif /* PROGRAM_VERSION */

#ifndef IV_LENGTH
#define IV_LENGTH 10 /* Ciphersaber compatible */
#endif /* IV_LENGTH */

typedef enum
{
	M_UNSPECIFIED,
	M_ENCRYPT, 
	M_DECRYPT
} action_t;

/* Prototypes */
void usage(void);
void version(void);
int get_random_bytes(unsigned char *buf, int len);
void encrypt_file(int input, int output, unsigned char *key_buf, int len);
void decrypt_file(int input, int output, unsigned char *key_buf, int len);
void open_files(const char *input_file, const char *output_file, 
					 const char *key_file, int *input, int *output, 
					 int *key);


void usage(void)
{
	fprintf(stderr, "Usage: %s [options]\n", PROGRAM_NAME);
	fprintf(stderr, "\n");
	fprintf(stderr, "A lightweight progem to encrypt and decrypt files using\n");
	fprintf(stderr, "the RC4 algorithm.\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "--encrypt, -E                Encrypt a file.\n");
	fprintf(stderr, "--decrypt, -D                Decrypt a file.\n");
	fprintf(stderr, "--input-file, -i  [file]     Specify input file.\n");
	fprintf(stderr, "--output-file, -o [file]     Specify output file.\n");
	fprintf(stderr, "--key-file, -k    [file]     Specify key file (max 246 bytes).\n");
	fprintf(stderr, "--help, -h                   Display this help text.\n");
	fprintf(stderr, "--version, -v                Display version and copyright information.\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "Instead of a filename, \"-\" can be specified in the \n");
	fprintf(stderr, "--input-file, --output-file and --key-file options to represent\n");
	fprintf(stderr, "standard input or standard output. Note that you cannot read both\n");
	fprintf(stderr, "the key and the input file from standard input simultaneously.\n");
	fprintf(stderr, "\n");
}

void version(void)
{
	fprintf(stderr, "%s %s\n", PROGRAM_NAME, PROGRAM_VERSION);
	fprintf(stderr, "Copyright 1999 Damien Miller <djm@mindrot.org>\n");
	fprintf(stderr, "http://www.mindrot.org/code/nanocrypt.php3\n");
	fprintf(stderr, "This program is licensed under the GNU GPL.\n");
}

int get_random_bytes(unsigned char *buf, int len)
{
	int urandom;
	int c;
	
	urandom = open("/dev/urandom", O_RDONLY);
	if (urandom == -1)
	{
		fprintf(stderr, "Couldn't open /dev/urandom: %s", strerror(errno));
		return(-1);
	}
	
	c = read(urandom, buf, len);
	if (c == -1)
	{
		fprintf(stderr, "Couldn't read from /dev/urandom: %s", strerror(errno));
		return(-1);
	}
	if (c != len)
	{
		fprintf(stderr, "Short read from /dev/urandom");
		return(-1);
	}

	return(0);
}

void encrypt_file(int input, int output, unsigned char *key_buf, int len)
{
	unsigned char *key_and_iv;
	rc4_t r;
	int bytes_read;
	int bytes_written;
	unsigned char buffer[8192];
	
	/* Make local copy of key with room for IV */
	key_and_iv = malloc(len + IV_LENGTH);
	if (key_and_iv == NULL)
	{
		fprintf(stderr, "Out of memory.\n");
		memset(key_buf, '\0', len);
		exit(1);
	}
	memcpy(key_and_iv, key_buf, len);

	/* Clean original key */
	memset(key_buf, '\0', len);

	/* Append IV */
	if (get_random_bytes(key_and_iv + len, IV_LENGTH) == -1)
	{
		/* Clean up */
		memset(key_and_iv, '\0', len + IV_LENGTH);
		exit(1);
	}

	/* Seed RC4 */
	rc4_key(&r, key_and_iv, len + IV_LENGTH);
	
	/* Write IV to output */
	bytes_written = write(output, key_and_iv + len, IV_LENGTH);

	/* Clear local copy of key + IV */
	memset(key_and_iv, '\0', len + IV_LENGTH);
	
	/* Check for errors in write */
	if (bytes_written != IV_LENGTH)
	{
		if (bytes_written == -1)
			fprintf(stderr, "Couldn't write to output file: %s\n", strerror(errno));
		else
			fprintf(stderr, "Short write to output file\n");
		
		exit(1);
	}
	
	/* Main loop: read bytes from input */
	while ((bytes_read = read(input, buffer, sizeof(buffer))) > 0)
	{
		/* Encrypt */
		rc4_crypt(&r, buffer, bytes_read);
		
		/* Write to output */
		bytes_written = write(output, buffer, bytes_read);
		if (bytes_written != bytes_read)
		{
			/* Clean up */
			memset(buffer, '\0', sizeof(buffer));
			memset(&r, '\0', sizeof(r));
			
			if (bytes_written == -1)
				fprintf(stderr, "Couldn't write to output file: %s\n", strerror(errno));
			else
				fprintf(stderr, "Short write to output file\n");

			exit(1);
		}
	}

	/* Clean up */
	memset(buffer, '\0', sizeof(buffer));
	memset(&r, '\0', sizeof(r));

	if (bytes_read == -1)
	{
		fprintf(stderr, "Couldn't read from input file: %s\n", strerror(errno));
		exit(1);
	}
}

void decrypt_file(int input, int output, unsigned char *key_buf, int len)
{
	unsigned char *key_and_iv;
	rc4_t r;
	int bytes_read;
	int bytes_written;
	unsigned char buffer[8192];
	
	/* Make local copy of key with room for IV */
	key_and_iv = malloc(len + IV_LENGTH);
	if (key_and_iv == NULL)
	{
		fprintf(stderr, "Out of memory.\n");
		memset(key_buf, '\0', len);
		exit(1);
	}
	memcpy(key_and_iv, key_buf, len);

	/* Clean original key */
	memset(key_buf, '\0', len);

	/* Append IV */
	bytes_read = read(input, key_and_iv + len, IV_LENGTH);
	if (bytes_read != IV_LENGTH)
	{
		memset(key_and_iv, '\0', len + IV_LENGTH);
		
		if (bytes_read == -1)
			fprintf(stderr, "Couldn't read from input file: %s\n", strerror(errno));
		else
			fprintf(stderr, "Short read from input file\n");
		
		exit(1);
	}

	/* Seed RC4 */
	rc4_key(&r, key_and_iv, len + IV_LENGTH);
	
	/* Clear local copy of key + IV */
	memset(key_and_iv, '\0', len + IV_LENGTH);
	
	/* Main loop: read bytes from input */
	while ((bytes_read = read(input, buffer, sizeof(buffer))) > 0)
	{
		/* Encrypt */
		rc4_crypt(&r, buffer, bytes_read);
		
		/* Write to output */
		bytes_written = write(output, buffer, bytes_read);
		if (bytes_written != bytes_read)
		{
			/* Clean up */
			memset(buffer, '\0', sizeof(buffer));
			memset(&r, '\0', sizeof(r));
			
			if (bytes_written == -1)
				fprintf(stderr, "Couldn't write to output file: %s\n", strerror(errno));
			else
				fprintf(stderr, "Short write to output file\n");

			exit(1);
		}
	}

	/* Clean up */
	memset(buffer, '\0', sizeof(buffer));
	memset(&r, '\0', sizeof(r));

	if (bytes_read == -1)
	{
		fprintf(stderr, "Couldn't read from input file: %s\n", strerror(errno));
		exit(1);
	}
}

void open_files(const char *input_file, const char *output_file, 
					 const char *key_file, int *input, int *output, 
					 int *key)
{
	/* Validate options */
	if (input_file == NULL)
	{
		fprintf(stderr, "You must specify an input file.\n");
		exit(1);
	}

	if (output_file == NULL)
	{
		fprintf(stderr, "You must specify an output file.\n");
		exit(1);
	}

	if (key_file == NULL)
	{
		fprintf(stderr, "You must specify a key file.\n");
		exit(1);
	}
	
	if ((strcmp(input_file, "-") == 0) && (strcmp(key_file, "-") == 0))
	{
		fprintf(stderr, "You cannot specify both input data and key to come from\n");
		fprintf(stderr, "standard input.\n");
		exit(1);
	}

	/* Open files */
	if (strcmp(input_file, "-") == 0)
	{
		*input = STDIN_FILENO;
	} else
	{
		*input = open(input_file, O_RDONLY);
		if (*input == -1)
		{
			fprintf(stderr, "Couldn't open input file: %s\n", strerror(errno));
			exit(1);
		}
	}
	
	if (strcmp(key_file, "-") == 0)
	{
		*key = STDIN_FILENO;
	} else
	{
		*key = open(key_file, O_RDONLY);
		if (*input == -1)
		{
			fprintf(stderr, "Couldn't open key file: %s\n", strerror(errno));
			exit(1);
		}
	}
	
	if (strcmp(output_file, "-") == 0)
	{
		*output = STDOUT_FILENO;
	} else
	{
		*output = open(output_file, O_WRONLY|O_CREAT|O_TRUNC, 0600);
		if (*input == -1)
		{
			fprintf(stderr, "Couldn't open input file: %s\n", strerror(errno));
			exit(1);
		}
	}
}

int main(int argc, char **argv)
{
	extern char *optarg;
	struct option long_options[] =
	{
		{"encrypt",		0, NULL, 'E'},
		{"decrypt",		0, NULL, 'D'},
		{"key-file",	1, NULL, 'k'},
		{"input",		1, NULL, 'i'},
		{"output",		1, NULL, 'o'},
		{"help",			0, NULL, 'h'},
		{"version",		1, NULL, 'v'},
		{NULL, 0, NULL, 0}
	};
	char *key_file = NULL;
	char *input_file = NULL;
	char *output_file = NULL;
	action_t mode = M_UNSPECIFIED;

	int input;
	int output;
	int key;
	
	unsigned char key_buf[256 - IV_LENGTH];
	int key_length;

	int c;

	uid_t old_uid;

	if (argc == 1)
	{
		usage();
		exit(1);
	}

	/* Fetch commandline options */
	while (1)
	{
		c = getopt_long(argc, argv, "EDk:i:o:hv?", long_options, NULL);

		if (c == -1)
			break;
		
		switch (c)
		{
			case 'E':
				mode = M_ENCRYPT;
				break;
			
			case 'D':
				mode = M_DECRYPT;
				break;
			
			case 'i':
				input_file = strdup(optarg);
				break;
			
			case 'o':
				output_file = strdup(optarg);
				break;
			
			case 'k':
				key_file = strdup(optarg);
				break;
			
			case 'v':
				version();
				exit(0);
				
			case 'h':
			case '?':
				usage();
				exit(0);
				
			default:
				fprintf(stderr, "Invalid commandline options.\n\n");
				usage();
				exit(1);
		}
	}

	/* Prevent outselves from being swapped out */
	old_uid = getuid();
	if (-1 == setuid(0))
	{
		fprintf(stderr, "Couldn't setuid to root.\n");
		exit(1);
	}
	if (-1 == mlockall(MCL_CURRENT|MCL_FUTURE))
	{
		fprintf(stderr, "Couldn't lock down pages.\n");
		exit(1);
	}
	if (-1 == setuid(old_uid))
	{
		fprintf(stderr, "Couldn't setuid to user.\n");
		exit(1);
	}

	open_files(input_file, output_file, key_file, &input, &output, &key);
	
	key_length = read(key, key_buf, sizeof(key_buf));
	if (key_length == -1)
	{
		fprintf(stderr, "Couldn't read key: %s\n", strerror(errno));
		exit(1);
	}
	key_length = (key_length < sizeof(key_buf))?key_length:sizeof(key_buf);

	/* Perform requested action */
	switch (mode)
	{
		case M_ENCRYPT:
			encrypt_file(input, output, key_buf, key_length);
			break;
		case M_DECRYPT:
			decrypt_file(input, output, key_buf, key_length);
			break;
		case M_UNSPECIFIED:
			fprintf(stderr, "Please select encryption (--encrypt) or decryption (--decrypt).\n");
			exit(1);
	}

	memset(key_buf, '\0', sizeof(key_buf));
	
	exit(0);
}
